import unittest

#import sys
#sys.path.append('../')

from handmadeML.optimizers.adam import *


#---------------------------------------------------------------------------
## tests for function NAME
#---------------------------------------------------------------------------
class Test_name_of_group(unittest.TestCase):

  def test_name_of_test(self):
    expected = 'value'
    actual = 'function(arguments)'
    self.assertTrue((actual != expected),\
            msg="message")



'''This module contains all functions needed for
the decision trees'''

#import pandas as pd
import numpy as np

## Functions

def node_construction(subtree,subX,suby,max_depth,min_samples_split):
    '''This recursive function is called by the fit method of
    dicision tree classifier

    Function for building the building the next left and right
    subnodes/leafs (and the associated subsplits of the dataset,
    depending of the current node and the associated dataset'''
    splits={}
    for index,col in subX.iteritems():
        splits[index]=np.sort(col.sample(1000,replace=True).unique())
        splits[index]=((splits[index][1:]+splits[index][:-1])/2).tolist()
    splits=[[(key,threshold) for threshold in value] for key,value in splits.items()]
    splits=[y for x in splits for y in x]

    ginis=[]
    ginis_left=[]
    ginis_right=[]
    counts_left=[]
    counts_right=[]
    for split in splits:
        mask=subX[split[0]]<split[1]
        y_left=suby[mask]
        y_right=suby[-mask]
        count_left=y_left.count()
        counts_left.append(count_left)
        count_right=y_right.count()
        counts_right.append(count_right)
        gini_left =1-((y_left.groupby( y_left).count()/ count_left)**2 ).sum()
        ginis_left.append(gini_left)
        gini_right=1-((y_right.groupby(y_right).count()/count_right)**2).sum()
        ginis_right.append(gini_right)
        gini_weighted= (y_left.count()*gini_left + y_right.count()*gini_right)/subtree['samples_count']
        ginis.append(gini_weighted)
    print(splits)
    print(ginis)

    best_split_index=ginis.index(min(ginis))
    best_split=splits[best_split_index]
    subtree['feature']=best_split[0]
    subtree['threshold']=best_split[1]
    left_gini  =ginis_left  [best_split_index]
    right_gini =ginis_right [best_split_index]
    left_count =counts_left [best_split_index]
    right_count=counts_right[best_split_index]
    mask=subX[best_split[0]]<best_split[1]
    y_left=suby[mask]
    y_right=suby[-mask]

    if left_gini < subtree['gini'] and\
      left_count >= min_samples_split and\
      subtree['level'] < max_depth and\
      left_gini > 0:
        X_left=subX[mask]
        subtree['left'] = {'level':subtree['level']+1,
                           'type':'node',
                           'gini':left_gini,
                           'samples_count':left_count,
                           'feature':None,
                           'threshold':None,
                           'left':None,
                           'right':None,
                           }
        node_construction(subtree['left'], X_left, y_left,     #recursion
                       max_depth=max_depth,
                       min_samples_split=min_samples_split)
    else:
        subtree['left'] = {'level':subtree['level']+1,
                           'type':'leaf',
                           'gini':left_gini,
                           'samples_count':left_count,
                           'category':y_left.mode()}

    if right_gini < subtree['gini'] and\
      right_count >= min_samples_split and\
      subtree['level'] < max_depth and\
      right_gini > 0:
        X_right=subX[-mask]
        subtree['right'] = {'level':subtree['level']+1,
                           'type':'node',
                           'gini':right_gini,
                           'samples_count':right_count,
                           'feature':None,
                           'threshold':None,
                           'left':None,
                           'right':None,
                           }
        node_construction(subtree['right'], X_right, y_right,   #recursion
                       max_depth=max_depth,
                       min_samples_split=min_samples_split)
    else:
        subtree['right'] = {'level':subtree['level']+1,
                           'type':'leaf',
                           'gini':right_gini,
                           'samples_count':right_count,
                           'category':y_right.mode()[0]}

def node_navigation(subtree, X_sample):
    '''This recursive function is called by the decision tree
    predict method
    X_sample : pd series (one single sample) with features as index'''
    if subtree['type']=='leaf':
        return subtree['category']
    feature=subtree['feature']
    if feature not in X_sample.index:
        raise ValueError(f'tree grown using feature {feature},\
which is not found in the columns of X')
    if X_sample[feature]>subtree['threshold']:
        return node_navigation(subtree['right'], X_sample)   #recursion
    return node_navigation(subtree['left'], X_sample)        #recursion



import unittest

import sys
sys.path.append('../')

from handmadeML.common.utils import *


#---------------------------------------------------------------------------
### tests for function compute_metric - normal mode (not derivative)########
#---------------------------------------------------------------------------
class Test_compute_metric(unittest.TestCase):

  def test_y_dimension_too_high_exception(self):
    test=unittest.TestCase()
    with test.assertRaises(ValueError) as context:
        compute_metric(np.array([[[1,1],[1,2]],
                                 [[2,1],[2,2]]]),
                       np.array([[1,2],
                                 [3,4]]),
                       'mse')
    self.assertTrue ('y vector dimension too high' in str(context.exception),\
        "no or wrong Exception raised when inputing a 3D-array as y\
         in compute_metric function")

#---------------------------------------------------------------------------
  def test_y_pred_dimension_too_high_exception(self):
    test=unittest.TestCase()
    with test.assertRaises(ValueError) as context:
        compute_metric(np.array([[1,2],
                                 [3,4]]),
                       np.array([[[1,1],[1,2]],
                                 [[2,1],[2,2]]]),
                       'mse')
    self.assertTrue ('y_pred vector dimension too high' in str(context.exception),\
        "no or wrong Exception raised when inputing a 3D-array as y_pred\
         in compute_metric function")

#---------------------------------------------------------------------------
  def test_unconsistent_dimensions_exception(self):
    test=unittest.TestCase()
    with test.assertRaises(ValueError) as context:
        compute_metric(np.array([[1,2,3],
                                 [4,5,6]]),
                       np.array([[1,2],
                                 [3,4]]),
                       'mse')
    self.assertTrue ('unconsistent vectors dimensions' in str(context.exception),\
        "no or wrong Exception raised when inputing unconsistent\
         y vs y_pred vectors shapes in compute_metric function")

#---------------------------------------------------------------------------
  def test_mse(self):
    self.assertTrue (compute_metric([1,0],[0.5,1],'mse') == 0.625,\
        "uncorrect mse metric behaviour")

#---------------------------------------------------------------------------
  def test_mse_multi_features(self):
    self.assertTrue (compute_metric([[1,0],[0,0]],[[0.5,1],[1,1]],'mse') == 0.8125,\
        "uncorrect mse metric behaviour for multi-features regressions\
         (2D y and y_pred vectors)")

#---------------------------------------------------------------------------
  def test_mae(self):
    self.assertTrue (compute_metric([1,0],[0.5,1],'mae') == 0.75,\
        "uncorrect mae metric behaviour")

#---------------------------------------------------------------------------
  def test_categorical_crossentropy(self):
    self.assertTrue (np.round(compute_metric([[1,0,0],[0,1,0]],[[0.8,0.1,0.1],[0.2,0.6,0.2]],
                                   'categorical_crossentropy'),
                    decimals=8) == 0.36698459,\
        "uncorrect categorical_crossentropy metric behaviour")

#---------------------------------------------------------------------------
  def test_binary_crossentropy(self):
    self.assertTrue (np.round(compute_metric([1,0],[0.9,0.1],'binary_crossentropy'),
                    decimals=8) == 0.10536052,\
        "uncorrect binary_crossentropy metric behaviour")

#---------------------------------------------------------------------------
  def test_y_dimension_too_high_with_binary_crossentropy_exception(self):
    test=unittest.TestCase()
    with test.assertRaises(ValueError) as context:
        compute_metric([[1,0,1],[0,0,0]],
                       [[0.5,0.9,0.1],
                        [0.9,0.9,0.1]],
                       'binary_crossentropy')
    self.assertTrue ('1 max for binary_crossentropy' in str(context.exception),\
        "no or wrong Exception raised when inputing 2D y/y_pred vectors\
         with binary_crossentropy selected in compute_metric function")

#---------------------------------------------------------------------------
  def test_unknown_metric_exception(self):
    test=unittest.TestCase()
    with test.assertRaises(ValueError) as context:
        compute_metric([0],[0],'typo_error')
    self.assertTrue ('Unknown metric' in str(context.exception),\
        "no or wrong Exception raised when inputing\
         unknown metric in compute_metric function")


#---------------------------------------------------------------------------
### tests for function compute_metric - derivative mode ####################
#---------------------------------------------------------------------------
class Test_metric_derivative(unittest.TestCase):

  def test_output_format(self):
    self.assertTrue (len(compute_metric([1,0],[0.5,1],'mse', loss_derivative = True)\
               .shape) == 2,\
        "uncorrect output : compute_metric must return a 2D array in derivative mode")

#---------------------------------------------------------------------------
  def test_mse(self):
    self.assertTrue ((compute_metric([1,0],[0.5,1],'mse', loss_derivative = True)\
                == np.array([[-0.5],[1]])).all(),\
        "uncorrect mse metric behaviour in derivative mode")

#---------------------------------------------------------------------------
  def test_mse_multi_features(self):
    self.assertTrue ((compute_metric([[1,0],[0,0]],[[0.5,1],[1,1]],'mse',
                       loss_derivative = True)\
                == np.array([[-0.25, 0.5],[0.5, 0.5]])).all(),\
        "uncorrect mse metric behaviour for multi-features regressions\
         (2D y and y_pred vectors) in derivative mode")

#---------------------------------------------------------------------------
  def test_mae(self):
    self.assertTrue ((compute_metric([1,0],[0.5,1],'mae', loss_derivative = True)\
                == np.array([[-0.5],[0.5]])).all(),\
        "uncorrect mae metric behaviour in derivative mode")

#---------------------------------------------------------------------------
  def test_categorical_crossentropy(self):
    self.assertTrue ((np.round(compute_metric([[1,0,0],[0,1,0]],[[0.8,0.1,0.1],[0.2,0.6,0.2]],
                                   'categorical_crossentropy',
                                   loss_derivative = True),
                    decimals=8) == np.array([[-0.625, -0.        , -0.],
                                             [-0.   , -0.83333333, -0.]])).all(),\
        "uncorrect categorical_crossentropy metric behaviour in derivative mode")

#---------------------------------------------------------------------------
  def test_binary_crossentropy(self):
    self.assertTrue ((np.round(compute_metric([1,0],[0.9,0.1],
                                   'binary_crossentropy',
                                   loss_derivative = True),
                    decimals=8) == np.array([[-0.55555556],
                                             [ 0.55555556]])).all(),\
        "uncorrect binary_crossentropy metric behaviour in derivative mode")




import unittest

#import sys
#sys.path.append('../')

from handmadeML.neuralnets.models import *

#---------------------------------------------------------------------------
## tests for method add_dense_layer
#---------------------------------------------------------------------------
class Test_add_dense_layer(unittest.TestCase):

  def test_no_input_dim_exception(self):

    my_nn=HandmadeNN()
    test=unittest.TestCase()
    with test.assertRaises(ValueError) as context:
        my_nn.add_dense_layer(5,'relu')
    self.assertTrue ('Use set_input_dim before creating first layer'\
           in str(context.exception),\
        "no or wrong Exception raised when adding first layer\
         to a network without setting input_dim")

#---------------------------------------------------------------------------
  def test_unknown_activation_exception(self):
    my_nn=HandmadeNN(5)
    test=unittest.TestCase()
    with test.assertRaises(ValueError) as context:
        my_nn.add_dense_layer(10,'typo_error')
    self.assertTrue ('Unknown activation type' in str(context.exception),\
        "no or wrong Exception raised when inputing\
         an unknown activation_type while adding layer")


#---------------------------------------------------------------------------
## tests for method predict - normal mode (without intermediate states) ####
#---------------------------------------------------------------------------
class Test_predict_without_intermediate(unittest.TestCase):

  def test_no_input_dim_exception(self):
    my_nn=HandmadeNN(5)# Empty neural network : just a pass-through for 5-values inputs

    self.assertTrue(my_nn.predict([2,3,2,3,4]).shape == (1,5),\
        "list not supported as an input for predict")
#---------------------------------------------------------------------------
  def test_list_as_input(self):
    my_nn=HandmadeNN(5)# Empty neural network : just a pass-through for 5-values inputs

    self.assertTrue(my_nn.predict([[2,3,2,3,4],[-2,-1,1,3,4]]).shape == (2,5),\
        "list of list not supported as an input for predict")

#---------------------------------------------------------------------------
  def test_x_dimension_too_high_exception(self):
    my_nn=HandmadeNN(5)# Empty neural network : just a pass-through for 5-values inputs

    test=unittest.TestCase()
    with test.assertRaises(ValueError) as context:
        my_nn.predict(np.array([[[1,1],[1,2],[1,3],[1,4],[1,5]],
                                    [[2,1],[2,2],[2,3],[3,4],[3,5]]]))
    self.assertTrue ('X vector dimension too high' in str(context.exception),\
        "no or wrong Exception raised when inputing a 3D-array in predict method")

#---------------------------------------------------------------------------
  def test_unconsistent_n_features_exception(self):
    my_nn=HandmadeNN(5)# Empty neural network : just a pass-through for 5-values inputs

    test=unittest.TestCase()
    with test.assertRaises(ValueError) as context:
        my_nn.predict(np.array([[1,1],[1,2],[1,3],[1,4],[1,5]]))
    self.assertTrue ('Unconsistent number of features' in str(context.exception),\
        "no or wrong Exception raised when inputing a X\
         with unconsistant size vs. network input_dim in predict method")

#---------------------------------------------------------------------------
  def test_oneD_array_as_input(self):
    my_nn=HandmadeNN(5)
    my_nn.add_dense_layer(10, 'linear')
    my_nn.weights[0] = np.vstack((np.identity(5),np.zeros((5,5))))
    self.assertTrue(my_nn.predict(np.array([-2,-1,2,3,4])).shape == (1,10),\
        "1-D array not supported as an input for predict method")

#---------------------------------------------------------------------------
  def test_output_dimension(self):
    my_nn=HandmadeNN(5)
    my_nn.add_dense_layer(10, 'linear')
    my_nn.weights[0] = np.vstack((np.identity(5),np.zeros((5,5))))
    self.assertTrue (my_nn.predict(np.array([[-2,-1,2,3,4],
                                         [-12,-11,12,13,14]])).shape == (2,10),\
        "the shape of the prediction for a 2*5 X input\
         by a network having 10neurons on last layer should be 2*10")

#--General-test-of-predict-method-with-all-activation-types-----------------
  def test_general_predict_test_all_activations(self):

    my_nn=HandmadeNN(5)

    my_nn.add_dense_layer(10, 'relu')
    my_nn.weights[-1] = np.concatenate([np.identity(5), np.zeros((5,5))], axis=0)
    my_nn.bias[-1] = np.array([0,0,0,0,1,1,1,0,0,0])

    my_nn.add_dense_layer(10, 'linear')
    my_nn.weights[-1] = np.flip(np.identity(10), 1)
    my_nn.bias[-1] = np.array([1,1,1,1,1,1,0,0,0,0])

    my_nn.add_dense_layer(10, 'tanh')
    my_nn.weights[-1] = np.identity(10)
    my_nn.bias[-1] = np.array([0,0,0,0,1,1,1,1,0,0])

    my_nn.add_dense_layer(10, 'softmax')
    my_nn.weights[-1] = np.flip(np.identity(10), 1)
    my_nn.bias[-1] = np.array([0,0,0,0,0,0,1,1,1,1])

    my_nn.add_dense_layer(1, 'sigmoid')
    my_nn.weights[-1] = np.expand_dims(np.arange(1,11,1), axis=0)
    my_nn.bias[-1] = np.array([0.5])


    self.assertTrue (np.round(my_nn.predict([-2,-1,2,3,4])[0,0], decimals=8) == 0.99939824,\
        "the general test of predict method on a network involving\
         all activation types and manually set bias and weights\
         did not return the correct value")


#---------------------------------------------------------------------------
## tests for method predict - backprop mode (with intermediate states) #####
#---------------------------------------------------------------------------
class Test_predict_backprop_mode(unittest.TestCase):

  def test_output_type_is_tuple(self):

    my_nn = HandmadeNN(5)# Empty neural network : just a pass-through for 5-values inputs
    my_nn.add_dense_layer(3, 'linear', weights_initializer='ones')

    outputs = my_nn.predict([2,3,2,3,4], keep_hidden_layers=True)
    output_type = type(outputs)
    self.assertTrue (output_type == tuple,\
        f"predict method with keep_hidden_layers must return a tuple. type is {output_type}")

#---------------------------------------------------------------------------
  def test_number_of_outputs(self):
    my_nn = HandmadeNN(5)# Empty neural network : just a pass-through for 5-values inputs
    my_nn.add_dense_layer(3, 'linear', weights_initializer='ones')

    outputs = my_nn.predict([2,3,2,3,4], keep_hidden_layers=True)
    n_outputs = len(outputs)

    self.assertTrue (n_outputs == 2,\
        f"predict method with keep_hidden_layers must return 2 output. here returns {n_outputs}")

#---------------------------------------------------------------------------
  def test_length_layers_outputs(self):
    my_nn = HandmadeNN(5)# Empty neural network : just a pass-through for 5-values inputs
    my_nn.add_dense_layer(3, 'linear', weights_initializer='ones')

    outputs = my_nn.predict([2,3,2,3,4], keep_hidden_layers=True)
    layers_outputs, layers_derivatives = outputs

    self.assertTrue (len(layers_outputs) == 2,\
        "the list of outputs of layers has not correct length using predict with keep_hidden_layers")

#---------------------------------------------------------------------------
  def test_length_layers_derivatives(self):
    my_nn = HandmadeNN(5)# Empty neural network : just a pass-through for 5-values inputs
    my_nn.add_dense_layer(3, 'linear', weights_initializer='ones')

    outputs = my_nn.predict([2,3,2,3,4], keep_hidden_layers=True)
    layers_outputs, layers_derivatives = outputs

    self.assertTrue (len(layers_derivatives) == 1,\
        "the list of derivatives of layers has not correct length using predict with keep_hidden_layers")

#---------------------------------------------------------------------------
  def test_layers_outputs_values(self):
    my_nn = HandmadeNN(5)# Empty neural network : just a pass-through for 5-values inputs
    my_nn.add_dense_layer(3, 'linear', weights_initializer='ones')

    outputs = my_nn.predict([2,3,2,3,4], keep_hidden_layers=True)
    layers_outputs, layers_derivatives = outputs

    self.assertTrue ((layers_outputs[1] == np.array([[14., 14., 14.]])).all(),\
        "uncorrect layers_outputs of predict method with keep_hidden_layers")

#---------------------------------------------------------------------------
  def test_layers_derivatives_values(self):
    my_nn = HandmadeNN(5)# Empty neural network : just a pass-through for 5-values inputs
    my_nn.add_dense_layer(3, 'linear', weights_initializer='ones')

    outputs = my_nn.predict([2,3,2,3,4], keep_hidden_layers=True)
    layers_outputs, layers_derivatives = outputs

    self.assertTrue ((layers_derivatives[0] == np.array([[1., 1., 1.]])).all(),\
        "uncorrect layers_derivatives of predict method with keep_hidden_layers")


#---------------------------------------------------------------------------
### tests for backpropagation method (computing the gradient) ##############
#---------------------------------------------------------------------------
class Test_backpropagation(unittest.TestCase):

  def test_output_type_is_tuple(self):

    my_nn=HandmadeNN(input_dim = 2)
    my_nn.add_dense_layer(2, 'linear', weights_initializer='ones')
    my_nn.add_dense_layer(1, 'linear', weights_initializer='ones')
    my_nn.set_loss('mse')

    outputs = my_nn.compute_backpropagation(np.array([[1,2],[2,3],[3,4]]), np.array([4,5,6]))
    output_type = type(outputs)

    self.assertTrue (output_type == tuple,\
        f"backpropagation method must return a tuple. type is {output_type}")

#---------------------------------------------------------------------------
  def test_number_of_outputs(self):
    my_nn=HandmadeNN(input_dim = 2)
    my_nn.add_dense_layer(2, 'linear', weights_initializer='ones')
    my_nn.add_dense_layer(1, 'linear', weights_initializer='ones')
    my_nn.set_loss('mse')

    outputs = my_nn.compute_backpropagation(np.array([[1,2],[2,3],[3,4]]), np.array([4,5,6]))
    n_outputs = len(outputs)

    self.assertTrue (n_outputs == 2,\
        f"backpropagation method must return 2 output. here returns {n_outputs}")

#---------------------------------------------------------------------------
  def test_weights_gradient_length(self):
    my_nn=HandmadeNN(input_dim = 2)
    my_nn.add_dense_layer(2, 'linear', weights_initializer='ones')
    my_nn.add_dense_layer(1, 'linear', weights_initializer='ones')
    my_nn.set_loss('mse')

    outputs = my_nn.compute_backpropagation(np.array([[1,2],[2,3],[3,4]]), np.array([4,5,6]))
    gradient_weights, gradient_bias = outputs

    self.assertTrue (len(gradient_weights) == 2,\
        "using backpropagation: the list of weights has not correct length")

#---------------------------------------------------------------------------
  def test_bias_gradient_length(self):
    my_nn=HandmadeNN(input_dim = 2)
    my_nn.add_dense_layer(2, 'linear', weights_initializer='ones')
    my_nn.add_dense_layer(1, 'linear', weights_initializer='ones')
    my_nn.set_loss('mse')

    outputs = my_nn.compute_backpropagation(np.array([[1,2],[2,3],[3,4]]), np.array([4,5,6]))

    gradient_weights, gradient_bias = outputs

    self.assertTrue (len(gradient_bias) == 2,\
        "using backpropagation: the list of bias has not correct length")

#---------------------------------------------------------------------------
  def test_weights_gradient_values(self):
    my_nn=HandmadeNN(input_dim = 2)
    my_nn.add_dense_layer(2, 'linear', weights_initializer='ones')
    my_nn.add_dense_layer(1, 'linear', weights_initializer='ones')
    my_nn.set_loss('mse')

    outputs = my_nn.compute_backpropagation(np.array([[1,2],[2,3],[3,4]]), np.array([4,5,6]))

    gradient_weights, gradient_bias = outputs

    self.assertTrue ((gradient_weights[0] == np.array([[24., 34.],
                                             [24., 34.]])).all(),\
        "using backpropagation: uncorrect gradient with respect to weights")

#---------------------------------------------------------------------------
  def test_bias_gradient_values(self):
    my_nn=HandmadeNN(input_dim = 2)
    my_nn.add_dense_layer(2, 'linear', weights_initializer='ones')
    my_nn.add_dense_layer(1, 'linear', weights_initializer='ones')
    my_nn.set_loss('mse')

    outputs = my_nn.compute_backpropagation(np.array([[1,2],[2,3],[3,4]]), np.array([4,5,6]))

    gradient_weights, gradient_bias = outputs

    self.assertTrue ((gradient_bias[0] == np.array([10., 10.])).all(),\
        "using backpropagation: uncorrect gradient with respect to bias")


#---------------------------------------------------------------------------
### tests for fit method
#---------------------------------------------------------------------------
class Test_fit_method(unittest.TestCase):

  def test_convergence_on_trivial_regression(self):
    my_nn=HandmadeNN(input_dim = 2)
    my_nn.add_dense_layer(1, 'linear',)
    X= np.ones((10_000, 2))
    y= np.zeros((10_000,1))
    my_nn.fit(X,y, loss='mse',optimizer_type='sgd', batch_size=7, n_epochs=2)

    self.assertTrue (my_nn.score(X,y,'mse') < 0.5,\
        "fit method has not converged with build-in sgd optimizer on a trivial regression")

#---------------------------------------------------------------------------
### tests for adam optimizer
#---------------------------------------------------------------------------
class Test_fit_method_with_adam(unittest.TestCase):

  def test_convergence_on_trivial_regression(self):
    my_nn=HandmadeNN(input_dim = 2)
    my_nn.add_dense_layer(1, 'linear',)
    X= np.ones((10_000, 2))
    y= np.zeros((10_000,1))
    my_nn.fit(X,y, loss='mse',optimizer_type='adam', batch_size=7, n_epochs=2)

    self.assertTrue (my_nn.score(X,y,'mse') < 0.5,\
        f"not converged with adam optimizer on a trivial regression : loss={my_nn.score(X,y,'mse')}")

#---------------------------------------------------------------------------


print ('all tests successfully passed')



'''This module contains all functions needed for
the neural network model'''

#import pandas as pd
import numpy as np
#from progressbar.bar import ProgressBar

## Functions

def compute_hinge(x, y, w, b, loss_type='hinge', derivative=False):
    '''compute the hinge or squared_hinge loss
    returns scalar
    inputs :
        x: shape (n_sample, n_features)
        y: 1/0 binary classes, array or list'''

    #converting DataFrames, lists or lists of lists to nparray
    y = np.array(y) #converting 0/1 class into -1/1
    x = np.array(x)

    losses_using_minus_plus1=['hinge','squared_hinge']
    if loss_type in losses_using_minus_plus1:
        y = (y-0.5)*2

    if loss_type=='hinge':
        if not derivative:
            return np.maximum(0, 1-(np.dot(x,w)+b)*y).mean()

        hinge_values=np.expand_dims(1-y*(np.dot(x,w)+b),axis=1)
        positives=(hinge_values>0)
        dloss_dw=-(positives*(np.expand_dims(y,axis=1)*x)).mean(axis=0)
        dloss_db=-(positives*np.expand_dims(y,axis=1)).mean(axis=0)
        return np.concatenate((dloss_dw, dloss_db),axis=0)

    if loss_type=='squared_hinge':
        if not derivative:
            return np.square(np.maximum(0, 1-(np.dot(x,w)+b)*y)).mean()

        hinge_values=np.expand_dims(1-y*(np.dot(x,w)+b),axis=1)
        positives=(hinge_values>0)
        dloss_dw=-2*(positives*(np.expand_dims(y,axis=1)*hinge_values*x)).mean(axis=0)
        dloss_db=-2*(positives*np.expand_dims(y,axis=1)*hinge_values).mean(axis=0)
        return np.concatenate((dloss_dw, dloss_db),axis=0)

    raise ValueError(f'unknown loss_type {loss_type}')



def compute_penalty(w,penalty_type='l2', derivative=False):
    w=np.array(w)
    if penalty_type=='l2':
        if not derivative:
            return 0.5 * np.dot(w.T,w)
        return w

    if penalty_type=='l1':
        if not derivative:
            return np.absolute(w).sum()
        return np.divide(w, np.absolute(w))

    if penalty_type==None:
        if not derivative:
            return 0
        return np.zeros_like(w)

    raise ValueError(f'unknown penalty_type {penalty_type}')



def compute_cost(x, y, params, C, loss_type='hinge',
                 penalty_type='l2', derivative=False):
    '''y must be 0/1 binary classes'''
    y = np.array(y)
    x = np.array(x)
    params=np.array(params)
    w = params[:-1]
    b = params[-1]

    #raise errors for unconsistant inputs
    if len(y.shape) > 2:
        raise ValueError('y vect  or dimension too high. Must be 2 max')

    if y.shape[0] != x.shape[0]:
        raise ValueError(f'unconsistent vectors dimensions during loss_derivative: \
y.shape= {y.shape} and x.shape= {x.shape}')

    if x.shape[1] != w.size:
        raise ValueError(f'unconsistent vectors dimensions during loss_derivative: \
x.shape= {x.shape} and w.size= {w.size}')


    if derivative is False:
        return C * compute_hinge(x, y, w, b, loss_type, derivative=False)\
               + compute_penalty(w, penalty_type, derivative=False)
    return C * compute_hinge(x, y, w, b, loss_type, derivative=True)\
           + np.concatenate((compute_penalty(w, penalty_type, derivative=True),
                                  np.zeros((1,))),
                                 axis=0)


#---------------------------------------------------------------------------
## functions for results plot
#---------------------------------------------------------------------------


def compute_line_points(a,b,c):
    '''computes 2 points far away from origin on the line defined by equation
    a*x + b*y +c = 0
    returns : [x1,x2], [y1,y2]'''
    delta=4 * a**2 * b**2 -4 * (a**2 + b**2) * (c**2 - 10000 * b**2)
    xroot1= (-2*a*c +np.sqrt(delta)) / 2 /(a**2 + b**2)
    xroot2= (-2*a*c -np.sqrt(delta)) / 2 /(a**2 + b**2)
    yroot1=(-c-a*xroot1)/b
    yroot2=(-c-a*xroot2)/b
    return [xroot1,xroot2], [yroot1,yroot2]



import unittest

#import sys
#sys.path.append('../')

from handmadeML.neuralnets.utils import *


#---------------------------------------------------------------------------
## tests for function compute_activation
#---------------------------------------------------------------------------
class Test_compute_activation(unittest.TestCase):

  def test_activation_relu(self):
    expected = np.array([[0,0], [0, 1], [1, 3]])
    actual = compute_activation(np.array([[-1,0], [0, 1], [1, 3]]), 'relu')
    self.assertTrue((actual == expected).all(),\
            msg="uncorrect relu function behaviour")

#---------------------------------------------------------------------------
  def test_activation_linear(self):
    expected = np.array([[-1,0], [0, 1], [1, 3]])
    actual = compute_activation(np.array([[-1,0], [0, 1], [1, 3]]), 'linear')

    self.assertTrue((actual == expected).all(),\
            msg="uncorrect linear function behaviour")

#---------------------------------------------------------------------------
  def test_activation_sigmoid(self):
    expected = np.array([[0.26894142, 0.5       ],
                         [0.5       , 0.73105858],
                         [0.73105858, 0.95257413]])
    actual = compute_activation(np.array([[-1,0], [0, 1], [1, 3]]),
                                        'sigmoid')
    self.assertTrue((np.round(actual,decimals=8) == expected).all(),\
            msg="uncorrect sigmoid function behaviour")

#---------------------------------------------------------------------------
  def test_activation_tanh(self):
    expected = np.array([[-0.76159416,  0.        ],
                         [ 0.        ,  0.76159416],
                         [ 0.76159416,  0.99505475]])
    actual = compute_activation(np.array([[-1,0], [0, 1], [1, 3]]),
                                        'tanh')
    self.assertTrue((np.round(actual,decimals=8) == expected).all(),\
            msg="uncorrect tanh function behaviour")

#---------------------------------------------------------------------------
  def test_activation_softmax(self):
    expected = np.array([[0.09003057, 0.04201007],
                         [0.24472847, 0.1141952 ],
                         [0.66524096, 0.84379473]])
    actual = compute_activation(np.array([[-1,0], [0, 1], [1, 3]]),
                                        'softmax')
    self.assertTrue((np.round(actual,decimals=8) == expected).all(),\
            msg="uncorrect softmax function behaviour")

#---------------------------------------------------------------------------
  def test_unknown_activation_type_error(self):
    test=unittest.TestCase()
    with test.assertRaises(ValueError) as context:
        compute_activation(0,'typo_error')
    self.assertTrue ('Unknown activation type' in str(context.exception),\
        "no or wrong Exception raised when inputing an unknown activation_type\
         while calling compute_activation")


#---------------------------------------------------------------------------
## tests for function compute_activation_derivative
#---------------------------------------------------------------------------
class Test_compute_activation_derivative(unittest.TestCase):

  def test_activation_derivative_relu(self):
    expected = np.array([[0,0], [0, 1], [1, 1]])
    actual = compute_activation_derivative(np.array([[-1,0], [0, 1], [1, 3]]),
                                          'relu')
    self.assertTrue((actual == expected).all(),\
                    msg="uncorrect relu function derivative behaviour")

#---------------------------------------------------------------------------
  def test_activation_derivative_linear(self):
    expected = np.array([[1,1], [1, 1], [1, 1]])
    actual = compute_activation_derivative(np.array([[-1,0], [0, 1], [1, 3]]),
                                          'linear')
    self.assertTrue((actual == expected).all(),\
                    msg="uncorrect linear function derivative behaviour")

#---------------------------------------------------------------------------
  def test_activation_derivative_sigmoid(self):
    expected = np.array([[0.19661193, 0.25      ],
                         [0.25      , 0.19661193],
                         [0.19661193, 0.04517666]])
    actual = compute_activation_derivative\
                     (np.array([[0.26894142, 0.5       ],
                                [0.5       , 0.73105858],
                                [0.73105858, 0.95257413]]),
                      'sigmoid')
    self.assertTrue((np.round(actual,decimals=8) == expected).all(),\
            msg="uncorrect sigmoid function derivative behaviour")

#---------------------------------------------------------------------------
  def test_activation_derivative_tanh(self):
    expected = np.array([[0.41997434, 1.        ],
                         [1.        , 0.41997434],
                         [0.41997434, 0.00986604]])
    actual = compute_activation_derivative\
                     (np.array([[-0.76159416,  0.        ],
                                [ 0.        ,  0.76159416],
                                [ 0.76159416,  0.99505475]]),
                      'tanh')
    self.assertTrue((np.round(actual,decimals=8) == expected).all(),\
            msg="uncorrect tanh function derivative behaviour")

#---------------------------------------------------------------------------
  def test_activation_derivative_softmax(self):
    expected = np.array([[0.08192507, 0.04024522],
                         [0.18483645, 0.10115466],
                         [0.22269543, 0.13180518]])
    actual = compute_activation_derivative\
                     (np.array([[0.09003057, 0.04201007],
                                [0.24472847, 0.1141952 ],
                                [0.66524096, 0.84379473]]),
                      'softmax')
    self.assertTrue((np.round(actual,decimals=8) == expected).all(),\
            msg="uncorrect softmax function derivative behaviour")

#---------------------------------------------------------------------------
  def test_unknown_activation_exception(self):
    test=unittest.TestCase()
    with test.assertRaises(ValueError) as context:
        compute_activation_derivative(0,'typo_error')
    self.assertTrue ('Unknown activation type' in str(context.exception),\
        "no or wrong Exception raised when inputing an unknown activation_type\
         while calling compute_activation_derivative")




import unittest

#import sys
#sys.path.append('../')

from handmadeML.linearmodels.models import *


#---------------------------------------------------------------------------
## ttests for method predict
#---------------------------------------------------------------------------
class Test_predict(unittest.TestCase):

  def test_list_as_input(self):
    my_svc=LinearClassifier(5)
    self.assertTrue(my_svc.predict([2,3,2,3,4]).shape == (1,),\
            msg="lists should be supported as an input for predict")

#---------------------------------------------------------------------------
  def test_list_of_lists_as_input(self):
    my_svc=LinearClassifier(5)
    self.assertTrue(my_svc.predict([[2,3,2,3,4],[-2,-1,1,3,4]]).shape == (2,),\
            msg="lists of lists should be supported as an input for predict")

#---------------------------------------------------------------------------
  def test_x_dim_too_high_exception(self):
    my_svc=LinearClassifier(5)
    test=unittest.TestCase()
    with test.assertRaises(ValueError) as context:
        my_svc.predict(np.array([[[1,1],[1,2],[1,3],[1,4],[1,5]],
                                    [[2,1],[2,2],[2,3],[3,4],[3,5]]]))
    self.assertTrue('X vector dimension too high' in str(context.exception),\
        "no or wrong Exception raised when inputing a 3D-array in predict method")

#---------------------------------------------------------------------------
  def test_unconsistent_features_exception(self):
    my_svc=LinearClassifier(5)
    test=unittest.TestCase()
    with test.assertRaises(ValueError) as context:
        my_svc.predict(np.array([[1,1],[1,2],[1,3],[1,4],[1,5]]))
    self.assertTrue('Unconsistent numbers of features' in str(context.exception),\
        "no or wrong Exception raised when inputing a X\
with unconsistant size vs. model input_dim in predict method")

#---------------------------------------------------------------------------
  def test_1D_array_as_input(self):
    my_svc=LinearClassifier(5)
    self.assertTrue(my_svc.predict(np.array([-2,-1,2,3,4])).shape == (1,),\
        "1-D array should be supported as an input for predict method")

#--General-test-of-predict-method-------------------------------------------
  def test_predict_general_test(self):
    my_svc=LinearClassifier(2)
    my_svc.params=np.array([1,-1,0])
    self.assertTrue((my_svc.predict([[1,2],[2,1]]) == np.array([0,1])).all(),\
        "the general test of predict method on svc with\
manually set params did not return the correct value")



'''This module contains an optimizer, which should work for
 all models, unless explicitly mentioned in the optimizer or model doc'''

import numpy as np

class AdamOptimizer():
    '''Adam Optimzer object designed primarly for neuralnets
    (so for list of arrays of params) but usable by linear models,
    altrough adam was devoped for neuralnets'''
    def __init__(self, params, alpha_init=0.001, beta_1=0.9,
             beta_2=0.999, epsilon=1e-8):
        '''params :
        has to be a tuple containing lists of params ndarrays
        for example : (list_of_weights_matrices, list_of_bias_matrices)
        '''

        self.alpha_init=alpha_init
        self.beta_1=beta_1
        self.beta_2=beta_2
        self.epsilon=epsilon


        self.t=0
        #initializing first and second momentum
        self.m = [[np.zeros_like(w) for w in l] for l in params]
        self.v = self.m.copy()

    def get_update(self, gradient):
        self.t+=1
        alpha_t=self.alpha_init*np.sqrt(1-self.beta_2**self.t)/(1-self.beta_1**self.t)

        # updating 1st and 2nd momenta
        self.m=[[self.beta_1 * mm + (1-self.beta_1) * gg\
                   for mm, gg in zip(ml, gl)] for ml, gl in zip(self.m,gradient)]
        self.v=[[self.beta_2 * vv + (1-self.beta_2) * gg**2\
                   for vv, gg in zip(vl, gl)] for vl, gl in zip(self.v,gradient)]

        #computing the updates
        print(self.v[0][0])
        update = [[- alpha_t * mm / (np.sqrt(vv) + self.epsilon)\
                        for mm, vv in zip(ml, vl)] for ml, vl in zip(self.m,self.v)]
        return tuple(update)


class OldAdamOptimizer():
    '''adam optimizer object
    This object in instanciated by the .fit() method of the model class
        each time it is triggered
    Unlike in Keras, this object should not be instanciated by the user

    WARNING : so far, this implementation is specific for the neural network model class
    '''

    def __init__(self, weights, bias, alpha_init=0.001, beta_1=0.9,
             beta_2=0.999, epsilon=1e-8):

        self.alpha_init=alpha_init
        self.beta_1=beta_1
        self.beta_2=beta_2
        self.epsilon=epsilon

        self.t=0
        #initializing first and second momentum
        self.m_weights = [np.zeros_like(w) for w in weights]
        self.m_bias = [np.zeros_like(b) for b in bias]
        self.v_weights = self.m_weights.copy()
        self.v_bias = self.m_bias.copy()

    def get_update(self, gradient_weights, gradient_bias):
        '''computes the values to be added to weights and bias arrays at
        the end of the train step'''
        self.t+=1
        alpha=self.alpha_init*np.sqrt(1-self.beta_2**self.t)/(1-self.beta_1**self.t)

        # updating 1st and 2nd momenta
        self.m_weights=[self.beta_1 * m + (1-self.beta_1) * grad\
                   for m, grad in zip(self.m_weights, gradient_weights)]
        self.m_bias=[self.beta_1 * m + (1-self.beta_1) * grad\
                   for m, grad in zip(self.m_bias, gradient_bias)]
        self.v_weights=[self.beta_2 * v + (1-self.beta_2) * grad**2\
                   for v, grad in zip(self.v_weights, gradient_weights)]
        self.v_bias=[self.beta_2 * v + (1-self.beta_2) * grad**2\
                   for v, grad in zip(self.v_bias, gradient_bias)]

        #computing the updates
        weights_update = [- alpha * m / (np.sqrt(v) + self.epsilon)\
                                  for m, v in zip( self.m_weights, self.v_weights)]
        bias_update = [- alpha * m / (np.sqrt(v) + self.epsilon)\
                                  for m, v in zip( self.m_bias, self.v_bias)]

        return weights_update, bias_update




'''This module contains decision trees models'''

#import numpy as np
import pandas as pd

from .utils import node_construction, node_navigation
#from ..common.utils import compute_metric
#from ..optimizers.adam import AdamOptimizer

class DecisionTreeClassifier ():
    '''equivalent to sklearn with limited options:
    handles numerical data only
    criterion : gini only
    splitter  : best only
    max_depth : None only (nodes are expanded until
         all leaves are pure or until all leaves
         contain less than min_samples_split samples
    min_samples_split : integers only
         (fraction of n_samples not supported)
    min_samples_leaf : 1 only
         (only the min_samples_splits criterions
         stops the growth of the tree)
    min_weight_fraction_leaf : 0 only
    max_features :None only (all features)
    max_leaf_nodes : None only (unlimited)
    min_impurity_decrease : 0 only
    min_impurity_split : 0 only
    '''
    def __init__(self, max_depth=10000, min_samples_split=2):
        self.max_depth=max_depth
        self.min_samples_split=min_samples_split
        self.tree={'level':0,
                   'type':'root',
                   'gini':None,
                   'samples_count':None,
                   'feature':None,
                   'threshold':None,
                   'left':None,
                   'right':None,
                   }

    def fit(self, X_train, y_train):
        '''y_train : pd.Series. If DataFrame, the 1st column will be used as target
           X_train : DataFrame or 2D array'''
        if isinstance(y_train,pd.core.frame.DataFrame):
            y_train=y_train.iloc[:,0]
        X_train=pd.DataFrame(X_train)
        self.tree['gini']=1-((y_train.groupby(y_train).count()/y_train.count())**2).sum()
        self.tree['samples_count']=y_train.count()

        node_construction(self.tree, X_train, y_train,
                          max_depth=self.max_depth,
                          min_samples_split=self.min_samples_split)

    def predict(self, X_test):
        X_test=pd.DataFrame(X_test)
        return X_test.apply(lambda x: node_navigation(self.tree, x), axis=1)



'''This module contains the linear models'''

import numpy as np
import pandas as pd
import seaborn as sns

from .utils import compute_cost,compute_line_points
from ..common.utils import compute_metric


from ..optimizers.adam import AdamOptimizer


class LinearClassifier ():
    '''
    hand-made classifier supporting both svm (without kernel trick)
    and logistic_regression depending of the loss.
    So far the options available are:

        - param initializers : 'uniform'

        - loss functions : 'hinge', 'squared_hinge', 'logistic'

        - solver : SGD without momentum 'sgd', 'adam'
    '''
    def __init__ (self, input_dim=0, param_initializer='random_uniform'):
        self.input_dim=input_dim
        if param_initializer=='random_uniform':
            self.params=np.random.uniform(size=(input_dim + 1,))
        else:
            raise ValueError('Unknown param initializer type')
        self.cost_history=[]

    def set_input_dim (self, input_dim):
        self.input_dim = input_dim
        self.params=np.random.uniform(size=(input_dim + 1,))

    def predict (self, X, mode='binary'):
        '''
        If mode=='binary':
            Computes binary prediction : wether the point
            is above or bellow the hyperplane of the svm
        If mode=='distance':
            Computes the distance from each data point to the svm hyperplane,
            normalized by the half margin (i.e. that distance=1 is for a point
            lying on the upper boundary of the margin, -1 is for lower boundary)
        If mode=='proba':
            Computes the prediction in the form of proba using sigmoid of the distance

        Input X : list, list of lists, np array, pd DataFrame
               axis 0 = samples
               axis 1 = features

        Output = y_pred: 1D np-array
               axis 0 = samples
               values : 0/1 if mode=='binary'
                        distances, normalized by the half margin if mode=='distance'
                        float if mode=='proba'
           '''
        #converting DataFrames, lists or lists of lists to nparray
        X = np.array(X)
        w = self.params[:-1]
        b = self.params[-1]

        #deal with 1D input to forge a 1 * n_features 2D-array
        if len(X.shape) == 1:
            X = np.expand_dims(X, axis = 0)

        #raise errors for unconsistant input
        if len(X.shape) > 2:
            raise ValueError('X vector dimension too high. Must be 2 max')
        if X.shape[1] != self.input_dim:
            raise ValueError(f'Unconsistent numbers of features. \
The network input_dim is {self.input_dim}')

        #compute the prediction
        distance = np.dot(X,w)+b
        if mode=='binary':
            return (distance>0)*1
        if mode=='distance':
            return distance
        if mode=='proba':
            return 1/(1+np.exp(-distance))

        raise ValueError(f'unknown prediction mode {mode}')


    def score (self, X, y, metric):
        '''uses predict method, then compute_metric function
        TODO : for some metrics, such as roc_auc, we will need predict_proba'''

        y_pred=self.predict(X)
        return compute_metric(y, y_pred, metric)


    def fit (self, X, y, loss_type='hinge', penalty_type='l2', C=1,
             learning_rate=0.01, batch_size=1, n_epochs=10, verbose=1,
             optimizer_type='sgd', alpha_init=0.001, beta_1=0.9,
             beta_2=0.999, epsilon=1e-8):
        '''input X : 2D array or pd DataFrame
                axis 0 = samples
                axis 1 = features
        '''

        if optimizer_type == 'adam':
            optimizer = AdamOptimizer (([self.params]),
                                        alpha_init=alpha_init, beta_1=beta_1,
                                        beta_2=beta_2, epsilon=epsilon)

        X = np.array(X)
        y = np.array(y)
        n_samples = X.shape[0]
        n_minibatches_per_epoch = int((n_samples-1) / batch_size) + 1

        cost = compute_cost(X, y, self.params, C, loss_type,
                            penalty_type, derivative=False)

        if self.cost_history == []:
            self.cost_history.append(cost)
        if verbose>0:
            print(f'initial cost: {cost}')
            print(f'initial margin: {2/np.sqrt(np.square(self.params[:-1]).sum())}')

        for epoch_index in range (n_epochs):
            if verbose>1:
                print(f'beginning epoch n°{epoch_index + 1}')

            #progress_batches = ProgressBar()
            #for mini_batch_index in progress_batches(range(n_minibatches_per_epoch)):
            for mini_batch_index in range(n_minibatches_per_epoch):
                gradient = compute_cost(X[mini_batch_index * batch_size :\
                                          (mini_batch_index +1) * batch_size],
                                        y[mini_batch_index * batch_size :\
                                          (mini_batch_index +1) * batch_size],
                                        self.params,
                                        C,
                                        loss_type,
                                        penalty_type,
                                        derivative=True)

                if optimizer_type == 'sgd':
                    #compute the update directly
                    params_update = -learning_rate * gradient

                elif optimizer_type == 'adam':
                    #compute the update with the optimizer
                    params_update = optimizer.get_update(gradient)

                else:
                    raise ValueError(f'unsupported optimizer type {optimizer}')

                # updating weights and bias
                self.params = self.params + params_update

            cost = compute_cost(X, y, self.params, C, loss_type,
                                penalty_type, derivative=False)
            self.cost_history.append(cost)
            if verbose>1:

                print(f'end of epoch n°{epoch_index + 1}. cost: {cost}')

        if verbose==1:
            print(f'final cost: {cost}')
            print(f'final margin: {2/np.sqrt(np.square(self.params[:-1]).sum())}')


    def plot_cost_history(self):
        '''Make a simple plot cost versus epochs'''
        graph=sns.lineplot(x=range(len(self.cost_history)),y=self.cost_history)
        graph.set(xlabel="epochs", ylabel = "cost")


    def plot_2D_svc(self,x,y):
        '''only in 2D : x.shape=(n_samples,2)
        y.shape=(n_samples,)'''
        params=self.params
        x=np.array(x)
        y=np.array(y)

        if x.shape[1]!=2:
            raise ValueError(f'x.shape={x.shape}. Expected (n_sample,2)')
        if params.size!=3:
            raise ValueError(f'input_dim={params.size-1}. Expected 2')

        df=pd.DataFrame({'x1':x[:,0],'x2':x[:,1],'y':y})

        xline, yline = compute_line_points(params[0],params[1],params[2])
        xlineup, ylineup = compute_line_points(params[0],params[1],params[2]-1)
        xlinedown, ylinedown = compute_line_points(params[0],params[1],params[2]+1)

        g=sns.scatterplot(data=df,x='x1',y='x2',hue='y')
        sns.lineplot(x=xlinedown, y=ylinedown)
        sns.lineplot(x=xlineup, y=ylineup)
        sns.lineplot(x=xline, y=yline)
        g.set(xlim=(df.x1.min()-2,df.x1.max()+2))
        g.set(ylim=(df.x2.min()-2,df.x2.max()+2))



'''This module contains all functions needed for
the neural network model'''

#import pandas as pd
import numpy as np
#from progressbar.bar import ProgressBar

## Functions

def compute_activation (X, activation_type):
    '''Defining activation functions
     Takes a nparray or a single value
    # Returns in the same format

    For softmax : assuming that X.shape[0]== n_neurons,
        the axis0 of array X is used for computing the mean
    '''
    X=np.array(X)
    if activation_type == 'relu':
        return np.maximum(X,0)
    if activation_type == 'sigmoid':
        return 1/(1+np.exp(-X))
    if activation_type == 'tanh':
        return np.tanh(X)
    if activation_type == 'linear':
        return X
    if activation_type == 'softmax':
        exp_x = np.exp(X)
        return exp_x / exp_x.sum(axis=0)

    #raise error if unknown type
    raise ValueError(f'Unknown activation type {activation_type}.\
Supported types : linear, relu, sigmoid, tanh, softmax')


def compute_activation_derivative (layer_output, activation_type):
    '''Computes the derivative of the activation functions,
       depending of the outputs of the output of these functions
           nota : if occures that for each of the 5 basic activations,
           f'(X) can be expressed simply as a function of f(X)

           Takes a nparray or a single value
        # Returns in the same format
           '''
    X_output=np.array(layer_output)
    if activation_type == 'relu':
        return (X_output > 0).astype(int)
    if activation_type == 'linear':
        return np.ones(X_output.shape)
    if activation_type == 'sigmoid':
        return X_output - np.square(X_output)
    if activation_type == 'tanh':
        return 1 - np.square(X_output)
    if activation_type == 'softmax':
        return X_output - np.square(X_output)

    #raise error if unknown type
    raise ValueError(f'Unknown activation type {activation_type}.\
Supported types : linear, relu, sigmoid, tanh, softmax')



'''Setup'''

from setuptools import find_packages
from setuptools import setup

with open('requirements.txt') as f:
    content = f.readlines()
requirements = [x.strip() for x in content if 'git+' not in x]

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setup(name='handmadeML',
      version="1.1",
      author='Thomas Mathieu',
      author_email='thomas.mathieu@m4x.org',
      description="'Personal machine/deep learning implementation inspired by sklearn and keras'",
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/bdepebhe/handmadeML',
      packages=find_packages(),
      classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        ],
      python_requires='>=3.6',
      install_requires=requirements,
      test_suite='tests',
      # include_package_data: to install data from MANIFEST.in
      include_package_data=True,
      scripts=['scripts/handmadeML-run'],
      zip_safe=False)




'''This module contains the neural network
model class itself'''

import numpy as np
import seaborn as sns

from .utils import compute_activation,compute_activation_derivative
from ..common.utils import compute_metric


from ..optimizers.adam import AdamOptimizer


class HandmadeNN ():
    '''
    hand-made version of neural network
    so far, the possibilities are :

        - layers activation functions :
            'linear', 'relu', 'sigmoid', 'tanh', 'softmax'

        - weights initializers : 'ones', 'glorot_uniform'
        - bias initializers : 'zeros', 'ones'

        - loss functions :
            'mse', 'mae', 'binary_crossentropy', 'categorical_crossentropy'

        - solver :
            SGD without momentum
    '''
    def __init__ (self, input_dim=0):
        self.weights=[]
        self.bias=[]
        self.activation_types=[]
        self.input_dim=input_dim
        self.n_layers=0

        self.loss_history=[]

    def set_input_dim (self, input_dim):
        '''manually sets the input_dim attribute of the model instance'''
        self.input_dim = input_dim

    def set_loss (self, loss):
        '''manually sets the loss attribute of the model instance'''
        self.loss = loss

    def add_dense_layer (self, n_neurons, activation_type,
                         weights_initializer='glorot_uniform', bias_initializer='zeros'):
        '''add a dense (fully connected) layer of neurons to the model
        This initializes the weights and bias according to selected initializer type,
        wich are yet implemented directly here'''
        #check if the input_dim is set
        if self.input_dim == 0:
            raise ValueError('input_dim = 0 .\
Use set_input_dim before creating first layer')

        #get the size of the input os this layer
        if len(self.bias) == 0:
            previous_dim=self.input_dim
        else:
            previous_dim=(self.bias[-1].shape[0])

        #initialize the layer parameters
        if weights_initializer == 'ones':
            self.weights.append(np.ones((n_neurons, previous_dim)))
        elif weights_initializer == 'glorot_uniform':
            limit = np.square(6 / (n_neurons + previous_dim))
            self.weights.append(np.random.uniform(-limit, limit, size = (n_neurons, previous_dim)))
        else:
            raise ValueError(f'Unknown weights initializer {weights_initializer}.\
Supported types : ones, glorot_uniform')

        if bias_initializer == 'zeros':
            self.bias.append(np.zeros(n_neurons))
        elif bias_initializer == 'ones':
            self.bias.append(np.ones(n_neurons))
        else:
            raise ValueError(f'Unknown bias initializer {bias_initializer}.\
Supported types : zeros, ones')

        self.activation_types.append(activation_type)
        self.n_layers += 1

        #test the activation type
        compute_activation(0, activation_type)

    def predict (self, X, keep_hidden_layers=False):
        '''input X : list, list of lists, np array, pd DataFrame
               axis 0 = samples
               axis 1 = features

           ## IF keep_hidden_layers==False:
           output = y_pred: 2D np-array
               axis 0 = samples
               axis 1 = output features, depending of the size of last layer

           ## IF keep_hidden_layers==True:
           outputs = layers_outputs, layers_activation_derivatives
           -output1 = layers_outputs:
               list of 2D np-arrays of outputs of each layer
               len(list)=n_layers+1: 1st element = X itself
                                     last element = y_pred
               axis 0 = samples
               axis 1 = number of neurons of the layer
           -output2 = layers_activation_derivatives:
               list of 2D np-arrays of d_act/d_input of each layer
               len(list)=n_layers
               axis 0 = samples
               axis 1 = number of neurons of the layer
           '''
        #converting DataFrames, lists or lists of lists to nparray
        X = np.array(X)

        #deal with 1D inputs to forge a 1 * n_features 2D-array
        if len(X.shape) == 1:
            X = np.expand_dims(X, axis = 0)

        #raise errors for unconsistant inputs
        if len(X.shape) > 2:
            raise ValueError('X vector dimension too high. Must be 2 max')
        if X.shape[1] != self.input_dim:
            raise ValueError(f'Unconsistent number of features. \
The network input_dim is {self.input_dim}')

        #compute the prediction
        layers_outputs = [X]
        layers_activation_derivatives = []
        for layer_index, activation_type in enumerate(self.activation_types):
            activation_input = np.dot(self.weights[layer_index], X.T)\
                               + np.expand_dims(self.bias[layer_index], axis=1)
            X = compute_activation(activation_input, activation_type).T

            layers_outputs.append(X)
            layers_activation_derivatives.append(\
                compute_activation_derivative(X, activation_type))

        if keep_hidden_layers:
            return layers_outputs, layers_activation_derivatives
        return X

    def score (self, X, y, metric):
        '''use predict method, then compute_metric function'''
        y_pred=self.predict(X)
        return compute_metric(y, y_pred, metric)

    def compute_backpropagation (self, X, y):
        '''This method :
            - executes self.predict(X) WITH keep_hidden_layers
                to keep all intermediate outputs
            - executes compute_metric (y, y_pred, loss) WITH loss_derivative
            - for each layer from last to first : computes loss
              derivatives (aka gradient) with respect to bias and weights

            output 1 : gradient with respect to weights
               (list of 2D arrays
               len(list) = n_layers
               axis 0 = number of neurons of the layer
               axis 1 = number of neurons of the previous layer (or features in the input)
            output 2 : gradient with respect to bias
               (list of 1D arrays)
               len(list) = n_layers
               axis 0 = number of neurons of the layer
            '''
        delta_weights=[]
        delta_bias=[]

        # compute the outputs and the derivatives of each layer
        layers_outputs, layers_activation_derivatives\
                = self.predict(X, keep_hidden_layers = True)
        # compute d_loss/d_ypred
        dloss_doutput = compute_metric (y,
                                        layers_outputs[-1],
                                        self.loss,
                                        loss_derivative = True)
        for layer_index in range(self.n_layers-1, -1, -1):
            # compute d_loss/d_input of the layer
            dloss_dinput = dloss_doutput * layers_activation_derivatives[layer_index]

            # compute gradient with respect to weights and bias
            delta_weights.append(np.dot(dloss_dinput.T, layers_outputs[layer_index]))
            delta_bias.append(np.sum(dloss_dinput, axis=0))

            # update dloss_doutput for next propagation
            if layer_index > 0:
                dloss_doutput = np.dot (dloss_dinput, self.weights[layer_index])

        delta_weights.reverse()
        delta_bias.reverse()

        return delta_weights, delta_bias



    def fit (self, X, y, loss=None, learning_rate=0.01,
             batch_size=1, n_epochs=10, verbose=1,
             optimizer_type='sgd',
             alpha_init=0.001, beta_1=0.9,
             beta_2=0.999, epsilon=1e-8):
        '''input X : 2D array or pd DataFrame
                axis 0 = samples
                axis 1 = features
        '''
        if loss:
            self.loss=loss

        if optimizer_type == 'adam':
            optimizer = AdamOptimizer ((self.weights, self.bias),
                                        alpha_init=alpha_init, beta_1=beta_1,
                                        beta_2=beta_2, epsilon=epsilon)
            #print(optimizer.v)

        X = np.array(X)
        y = np.array(y)
        n_samples = X.shape[0]
        n_minibatches_per_epoch = int(n_samples / batch_size)

        loss=self.score(X, y, self.loss)
        if self.loss_history == []:
            self.loss_history.append(loss)

        if verbose>0:
            print(f'initial loss: {self.score(X, y, self.loss)}')

        for epoch_index in range (n_epochs):
            if verbose>1:
                print(f'beginning epoch n°{epoch_index + 1}')

            #progress_batches = ProgressBar()
            #for mini_batch_index in progress_batches(range(n_minibatches_per_epoch)):
            for mini_batch_index in range(n_minibatches_per_epoch):
                gradient_weights, gradient_bias\
                    = self.compute_backpropagation(X[mini_batch_index * batch_size :\
                                                     (mini_batch_index +1) * batch_size],
                                                   y[mini_batch_index * batch_size :\
                                                     (mini_batch_index +1) * batch_size])

                if optimizer_type == 'sgd':
                    #compute the update directly
                    weights_update = [-learning_rate * grad for grad in gradient_weights]
                    bias_update = [-learning_rate * grad for grad in gradient_bias]

                elif optimizer_type == 'adam':
                    #compute the update with the optimizer
                    weights_update, bias_update = optimizer.get_update((gradient_weights,
                                                                       gradient_bias))

                else:
                    raise ValueError(f'unsupported optimizer type {optimizer_type}')

                # updating weights and bias
                self.weights = [w + w_update  for w, w_update in zip(self.weights, weights_update)]
                self.bias = [b + b_update for b, b_update in zip(self.bias, bias_update)]

            loss=self.score(X, y, self.loss)
            self.loss_history.append(loss)

            if verbose>1:
                print(f'end of epoch n°{epoch_index + 1}. loss: {self.score(X, y, self.loss)}')
        if verbose==1:
            print(f'final loss: {self.score(X, y, self.loss)}')

    def plot_loss_history(self):
        '''plots the complete loss history of the model since creation,
        including multiple .fit() calls'''
        graph=sns.lineplot(x=range(len(self.loss_history)),y=self.loss_history)
        graph.set(xlabel="epochs", ylabel = "loss")



'''This module contains common utility functions'''

#import pandas as pd
import numpy as np
#from progressbar.bar import ProgressBar

## Functions

def compute_metric (y, y_pred, metric, loss_derivative=False):
    '''Defining loss and metric functions
     Takes nparrays, lists or a single values

     ## IF loss_derivative==False:
         output: always scalar

     ## IF loss_derivative==True: (True will be ignored for non-loss metrics)
         Computes the partial derivative of the loss function
           with respect to each component of each sample
         output: 2Darray
            n_samples * 1 for binary_crossentropy or single output regression
            n_samples * n_class for categorical_crossentropy
            n_samples * n_features for multifeatures regression)
    '''

    #converting DataFrames, lists or lists of lists to nparray
    y = np.array(y)
    y_pred = np.array(y_pred)

    #deal with 1D inputs to forge a n_samples * 1 2D-array
    if len(y.shape) == 1:
        y = np.expand_dims(y, axis = 1)
    if len(y_pred.shape) == 1:
        y_pred = np.expand_dims(y_pred, axis = 1)

    #raise errors for unconsistant inputs
    if len(y.shape) > 2:
        raise ValueError('y vector dimension too high. Must be 2 max')
    if len(y_pred.shape) > 2:
        raise ValueError('y_pred vector dimension too high. Must be 2 max')
    if y.shape != y_pred.shape:
        raise ValueError(f'unconsistent vectors dimensions during scoring :\
y.shape= {y.shape} and y_pred.shape= {y_pred.shape}')

    #compute loss funtions (or derivatives if loss_derivative==True)
    if metric == 'mse':
        if not loss_derivative:
            return np.square(y-y_pred).mean()
        return 1 / y.size * 2 * (y_pred - y)

    if metric == 'mae':
        if not loss_derivative:
            return np.abs(y-y_pred).mean()
        return 1 / y.size * (y_pred - y) / np.abs(y - y_pred)

    if metric == 'categorical_crossentropy':
        if not loss_derivative:
            return -1 / y.shape[0] * ((y * np.log(y_pred)).sum())
        return -1 / y.shape[0] * (y / y_pred)

    if metric == 'binary_crossentropy':
        if y.shape[1]>1:
            raise ValueError('y vector dimension too high.\
Must be 1 max for binary_crossentropy')
        if not loss_derivative:
            return -(y*np.log(y_pred)+(1-y)*np.log(1-y_pred)).mean()
        return -1 / y.size * (y / y_pred - (1-y) / (1-y_pred))

    # compute other metrics functions
    #### accuracy, f1-score, recall, etc.. : not implemented yet

    raise ValueError(f'Unknown metric {metric}. Supported types :\
mse, mae, categorical_crossentropy, binary_crossentropy')



import unittest

#import sys
#sys.path.append('../')
import pandas as pd


from handmadeML.decisiontrees.models import *


#---------------------------------------------------------------------------
## ttests for method predict
#---------------------------------------------------------------------------
class Test_predict(unittest.TestCase):

  def test_predict_method_behaviour(self):
    dtc=DecisionTreeClassifier(min_samples_split=10)
    dtc.tree={'level': 0,
             'type': 'root',
             'feature': 'aa',
             'threshold': 0.5,
             'left': None,
             'right': None}
    dtc.tree['left']={'level': 1,
                 'type': 'root',
                 'feature': 'bb',
                 'threshold': 0.5,
                 'left': {'level':2,'type':'leaf','category':1},
                 'right': {'level':2,'type':'leaf','category':0}}
    dtc.tree['right']={'level':1,
                  'type':'leaf',
                  'category':1}
    X_test=pd.DataFrame({'aa':[0.1,0.2,0.7],'bb':[1,0,1],'cc':[1,1,0]})
    prediction = dtc.predict(X_test)
    self.assertTrue((prediction==pd.Series([0,1,1])).all(),\
        f"uncorrect predict method. returned {prediction}, expected [0,1,1]")
